// 
// Interface Between the ATM Dialog and the Winsock2 ATM API
// Functions called by the CSettingsDlg Class :
//      - GetATMDeviceCount 
//      - SockGetLocalATMAddress
//      - SockStartATMConnection
//      - StopAtmConnection


#include "stdafx.h"
#include <setupapi.h>
#include "testatm.h"
#include "patmif.h"
#include "common.h"
#include "sockatm.h"
#include <ws2atm.h>	
#include "public.h"
#include <conio.h>


CONNECTION_INFO ConnectionInfo;

extern BOOL bStepMode;

static DWORD WaitCallThread(DWORD param);
static DWORD RecvThread(DWORD param);
static DWORD XmitThread(DWORD param);


// Initializes WinSock2
static BOOL InitWS2(void)
{
	WORD ver_req;
	WSADATA wsa_data;

    ver_req = MAKEWORD(2, 0);
	
	if (WSAStartup(ver_req, &wsa_data) != 0)
	{
		THROW_MESSAGE("Error : Starting up WinSock2.0 DLL",0);
		return FALSE;
	}

	if (LOBYTE(wsa_data.wVersion) != 2 || HIBYTE(wsa_data.wVersion) != 0)
	{
		THROW_MESSAGE("Error : Could not find WinSock2.0 DLL",0);
		WSACleanup();
		return FALSE;
	}

	return TRUE;

}

// Retreives the the VPI and VCI of a SVC connection
static BOOL GetATMConnectionId(SOCKET Socket,ATM_CONNECTION_ID *pConnID)
{
    ULONG Bytes;

	if (WSAIoctl(Socket, SIO_GET_ATM_CONNECTION_ID, NULL, 0, (LPVOID) pConnID, 

		         sizeof(ATM_CONNECTION_ID), &Bytes, NULL, NULL) == SOCKET_ERROR)
	{

		THROW_MESSAGE("Error: WSAIoctl(SIO_GET_ATM_CONNECTION_IDSocket. Error Code:%d", WSAGetLastError());
		return FALSE;
	}
    return TRUE;
}

// Initializes A SVC Connection
static BOOL CreateSVC(int DeviceNumber, int SvcType, PSVC_SETTINGS pSvcSettings, int QosType,
                      BOOL SetQos, PQOS_SETTINGS pQosSettings ,int PacketSize)
{


	struct sockaddr_atm SocketAddress;
    SOCKET Socket;
	int    AddressLength;
    ATM_CONNECTION_ID ConnID;

	AAL_PARAMETERS_IE                  AalParameters;
	ATM_TRAFFIC_DESCRIPTOR_IE          TrafficDescriptor;
	ATM_BROADBAND_BEARER_CAPABILITY_IE BroadbandBearerCapability;
	ATM_QOS_CLASS_IE                   QosClass;
	Q2931_IE                           *pQ2931;
    QOS                                Qos;

    int size;


    if (!InitWS2())
        return SOCKET_ERROR;

    if (SockGetLocalATMAddress(DeviceNumber,ConnectionInfo.LocalATMAddress) != TRUE)   
    {

	    WSACleanup();
 		return SOCKET_ERROR;

    }

	
    /* Create a socket */
    if ((Socket = WSASocket(AF_ATM, SOCK_RAW, ATMPROTO_AAL5, NULL, NO_GROUP, 
			        WSA_FLAG_OVERLAPPED)) == SOCKET_ERROR)
	{
		THROW_MESSAGE("Error: WSASocket. Error Code:%d", WSAGetLastError());
        WSACleanup();
		return SOCKET_ERROR;
	}


    // In case we are the caller, we must specify the transmit QOS settings for us ( Forward) and for 
    // the called party (Backward).
    if (SvcType == CONNECTION_SVC_OUT)
    {

 	// setting up Qos parameters ...

	    AalParameters.AALType = AALTYPE_5;
    	AalParameters.AALSpecificParameters.AAL5Parameters.ForwardMaxCPCSSDUSize  = PacketSize;
	    AalParameters.AALSpecificParameters.AAL5Parameters.BackwardMaxCPCSSDUSize = PacketSize;
	    AalParameters.AALSpecificParameters.AAL5Parameters.SSCSType               = AAL5_SSCS_NULL;

	    size = sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(AAL_PARAMETERS_IE);

	    TrafficDescriptor.Forward.PeakCellRate_CLP0         = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Forward.PeakCellRate_CLP01        = SAP_FIELD_ABSENT;
        TrafficDescriptor.Forward.SustainableCellRate_CLP0  = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Forward.SustainableCellRate_CLP01 = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Forward.MaxBurstSize_CLP0         = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Forward.MaxBurstSize_CLP01        = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Forward.Tagging                   = SAP_FIELD_ABSENT;
    

	    TrafficDescriptor.Backward.PeakCellRate_CLP0         = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Backward.PeakCellRate_CLP01        = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Backward.SustainableCellRate_CLP0  = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Backward.SustainableCellRate_CLP01 = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Backward.MaxBurstSize_CLP0         = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Backward.MaxBurstSize_CLP01        = SAP_FIELD_ABSENT;
	    TrafficDescriptor.Backward.Tagging                   = SAP_FIELD_ABSENT;


        if ( (QosType == QOS_UBR || QosType == QOS_CBR ) && SetQos)
        {
            TrafficDescriptor.Forward.PeakCellRate_CLP01 = pQosSettings->Ubr.PeakCellRate;
            TrafficDescriptor.Backward.PeakCellRate_CLP01 = pQosSettings->Ubr.PeakCellRate;
        }
  
        else if (QosType == QOS_VBR && SetQos)
        {
            TrafficDescriptor.Forward.PeakCellRate_CLP01        = pQosSettings->Vbr.PeakCellRate;
            TrafficDescriptor.Forward.SustainableCellRate_CLP01 = pQosSettings->Vbr.SubstainableCellRate;
            TrafficDescriptor.Forward.MaxBurstSize_CLP01        = pQosSettings->Vbr.MaxBurstSize;
        
            TrafficDescriptor.Backward.PeakCellRate_CLP01        = pQosSettings->Vbr.PeakCellRate;
            TrafficDescriptor.Backward.SustainableCellRate_CLP01 = pQosSettings->Vbr.SubstainableCellRate;
            TrafficDescriptor.Backward.MaxBurstSize_CLP01        = pQosSettings->Vbr.MaxBurstSize;
        }


        TrafficDescriptor.BestEffort = TRUE;         
        if ( QosType != QOS_UBR )
	        TrafficDescriptor.BestEffort = 0;       // Note: this must be set to zero for CBR and VBR

	    size += sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(ATM_TRAFFIC_DESCRIPTOR_IE);

	    BroadbandBearerCapability.BearerClass = BCOB_X;
    
        BroadbandBearerCapability.TrafficType = TT_NOIND;
        if ( QosType == QOS_CBR )
            BroadbandBearerCapability.TrafficType = TT_CBR;
        else if ( QosType == QOS_VBR )
            BroadbandBearerCapability.TrafficType = TT_VBR;

	    BroadbandBearerCapability.TimingRequirements = TR_END_TO_END;
	    BroadbandBearerCapability.ClippingSusceptability = CLIP_NOT;
	    BroadbandBearerCapability.UserPlaneConnectionConfig = UP_P2P;
    
        size += sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(ATM_BROADBAND_BEARER_CAPABILITY_IE);
  
        QosClass.QOSClassForward  = QOS_CLASS0;
        QosClass.QOSClassBackward = QOS_CLASS0;     // CLASS0 is UBR
    
        if ( QosType == QOS_CBR )
        {
            QosClass.QOSClassForward  = QOS_CLASS1;
            QosClass.QOSClassBackward = QOS_CLASS1; // CLASS1 is CBR
        }
        else if ( QosType == QOS_VBR )
        {

            QosClass.QOSClassForward  = QOS_CLASS2;
            QosClass.QOSClassBackward = QOS_CLASS2; // CLASS2 is VBR
        }

        size += sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(ATM_QOS_CLASS_IE);
    
	
        Qos.ProviderSpecific.buf = (char *) malloc(size);
        Qos.ProviderSpecific.len = size;
        memset(Qos.ProviderSpecific.buf, 0, size);

        pQ2931 = (Q2931_IE *) Qos.ProviderSpecific.buf;
        pQ2931->IEType = IE_AALParameters;
        pQ2931->IELength = sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(AAL_PARAMETERS_IE);
        memcpy(pQ2931->IE, &AalParameters, sizeof(AAL_PARAMETERS_IE));

        pQ2931 = (Q2931_IE *) ((char *)pQ2931 + pQ2931->IELength);
        pQ2931->IEType = IE_TrafficDescriptor;
        pQ2931->IELength = sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(ATM_TRAFFIC_DESCRIPTOR_IE);
        memcpy(pQ2931->IE, &TrafficDescriptor, sizeof(ATM_TRAFFIC_DESCRIPTOR_IE));

        pQ2931 = (Q2931_IE *) ((char *)pQ2931 + pQ2931->IELength);
        pQ2931->IEType = IE_BroadbandBearerCapability;
        pQ2931->IELength = sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(ATM_BROADBAND_BEARER_CAPABILITY_IE);
        memcpy(pQ2931->IE, &BroadbandBearerCapability, sizeof(ATM_BROADBAND_BEARER_CAPABILITY_IE));

	
        pQ2931 = (Q2931_IE *) ((char *)pQ2931 + pQ2931->IELength);
        pQ2931->IEType = IE_QOSClass;
        pQ2931->IELength = sizeof(Q2931_IE_TYPE) + sizeof(ULONG) + sizeof(ATM_QOS_CLASS_IE);
        memcpy(pQ2931->IE, &QosClass, sizeof(ATM_QOS_CLASS_IE));

        // since the QOS settings are specified in the ProviderSpecific field, 
        // the following fields are let to default
   	    Qos.SendingFlowspec.TokenRate           = QOS_NOT_SPECIFIED;
	    Qos.SendingFlowspec.TokenBucketSize     = QOS_NOT_SPECIFIED;
	    Qos.SendingFlowspec.PeakBandwidth       = QOS_NOT_SPECIFIED;
	    Qos.SendingFlowspec.Latency             = QOS_NOT_SPECIFIED;
	    Qos.SendingFlowspec.DelayVariation      = QOS_NOT_SPECIFIED;
	    Qos.SendingFlowspec.ServiceType         = SERVICETYPE_BESTEFFORT;
	    Qos.SendingFlowspec.MaxSduSize          = QOS_NOT_SPECIFIED;
	    Qos.SendingFlowspec.MinimumPolicedSize  = QOS_NOT_SPECIFIED;

        Qos.ReceivingFlowspec.TokenRate           = QOS_NOT_SPECIFIED;
	    Qos.ReceivingFlowspec.TokenBucketSize     = QOS_NOT_SPECIFIED;
	    Qos.ReceivingFlowspec.PeakBandwidth       = QOS_NOT_SPECIFIED;
	    Qos.ReceivingFlowspec.Latency             = QOS_NOT_SPECIFIED;
	    Qos.ReceivingFlowspec.DelayVariation      = QOS_NOT_SPECIFIED;
	    Qos.ReceivingFlowspec.ServiceType         = SERVICETYPE_BESTEFFORT;
	    Qos.ReceivingFlowspec.MaxSduSize          = QOS_NOT_SPECIFIED;
	    Qos.ReceivingFlowspec.MinimumPolicedSize  = QOS_NOT_SPECIFIED;


        memset((void *)&SocketAddress, 0, sizeof(struct sockaddr_atm));
	    SocketAddress.satm_family = AF_ATM;
	    SocketAddress.satm_number.AddressType = ATM_NSAP;
	    SocketAddress.satm_number.NumofDigits = 20;
	    SocketAddress.satm_blli.Layer2Protocol = SAP_FIELD_ANY; 
	    SocketAddress.satm_blli.Layer3Protocol = SAP_FIELD_ABSENT;
	    SocketAddress.satm_bhli.HighLayerInfoType = SAP_FIELD_ABSENT;
    

	    /* Connect to the SERVER with the address specified in the SocketAddress */

	    memcpy(SocketAddress.satm_number.Addr, pSvcSettings->ServerAddress, 20);
	
    	if (WSAConnect(Socket, (struct sockaddr FAR *)&SocketAddress,
    		sizeof(struct sockaddr_atm), NULL, NULL, &Qos, NULL) < 0) 
        {
		    THROW_MESSAGE("Error: WSAConnnect. Error Code:%d", WSAGetLastError());
		    closesocket(Socket);
	        WSACleanup();
            free(Qos.ProviderSpecific.buf);
		    return SOCKET_ERROR;
        }
        
        memcpy(ConnectionInfo.RemoteATMAddress,SocketAddress.satm_number.Addr,20);

        
        ConnID.DeviceNumber = DeviceNumber;
        if (GetATMConnectionId(Socket,&ConnID))
        {
            ConnectionInfo.Vpi=ConnID.VPI;
            ConnectionInfo.Vci=ConnID.VCI;
            ConnectionInfo.DeviceNumber= DeviceNumber;

        }
 
        free(Qos.ProviderSpecific.buf);

	}

    // in case we wait for a connection, it is the caller which specifies the Qos settings
    else
    {

         ConnectionInfo.DeviceNumber= DeviceNumber;

        if (NULL==(ConnectionInfo.hAccept=(WSAEVENT)WSACreateEvent()))       // create event handle for connect events
	    {
		    THROW_MESSAGE("Error: WSACreateEvent. Error Code:%d", GetLastError());
	        return SOCKET_ERROR;
	    }

        // set WSASelect for connect events
        if (SOCKET_ERROR==WSAEventSelect(Socket,ConnectionInfo.hAccept,FD_ACCEPT))
	    {
		    THROW_MESSAGE("Error: WSAEventSelect. Error Code:%d", WSAGetLastError());
	        return SOCKET_ERROR;
	    }


        memset((void *)&SocketAddress, 0, sizeof(struct sockaddr_atm));
	    SocketAddress.satm_family = AF_ATM;
	    SocketAddress.satm_number.AddressType = ATM_NSAP;
	    SocketAddress.satm_number.NumofDigits = 20;
	    SocketAddress.satm_blli.Layer2Protocol = SAP_FIELD_ANY; 
	    SocketAddress.satm_blli.Layer3Protocol = SAP_FIELD_ABSENT;
	    SocketAddress.satm_bhli.HighLayerInfoType = SAP_FIELD_ABSENT;
    
        memcpy(SocketAddress.satm_number.Addr,ConnectionInfo.LocalATMAddress,19);
    	SocketAddress.satm_number.Addr[ATM_ADDR_SIZE-1] = pSvcSettings->Selector;


	    if (bind(Socket, (struct sockaddr FAR *)&SocketAddress, sizeof(struct sockaddr_atm))== SOCKET_ERROR)
	    {
		    THROW_MESSAGE("Error: bind. Error Code:%d", WSAGetLastError());
		    closesocket(Socket);
	        WSACleanup();
 		    return SOCKET_ERROR;
	    }

	    if (listen(Socket, 5) == SOCKET_ERROR)
	    {
		    THROW_MESSAGE("Error: listen. Error Code:%d", WSAGetLastError());
		    closesocket(Socket);
	        WSACleanup();
 		    return SOCKET_ERROR;
        
        }
        AddressLength=sizeof(SocketAddress);

    }

    ConnectionInfo.PacketSize = PacketSize;	
    return Socket;
}

// Initializes a PVC Connection
static int CreatePVC( int DeviceNumber,
                      int Vpi, int Vci, 
                      int AalType ,
					  int Gfc, int Pt, int Clp,
                      int QosType ,BOOL SetQos,PQOS_SETTINGS pQosSettings,
                      int PacketSize)
{

	ATM_PVC_PARAMS      PvcParameters;
    DWORD               OutBytes;
    PFLOWSPEC           pSendQos, pRecvQos;
    int ProtocolType ;
    SOCKET Socket;

    if (!InitWS2())
        return SOCKET_ERROR;
    

    ProtocolType = ATMPROTO_AAL5;
    
    /* Create a AAL5- AALUSER  socket */
    if ((Socket = WSASocket(AF_ATM, SOCK_RAW, ProtocolType, NULL, NO_GROUP, 
			        WSA_FLAG_OVERLAPPED)) == SOCKET_ERROR)
	{
		THROW_MESSAGE("Error: WSASocket. Error Code:%d", WSAGetLastError());
        WSACleanup();
		return SOCKET_ERROR;
	}

    pRecvQos = &PvcParameters.PvcQos.ReceivingFlowspec;
    pSendQos = &PvcParameters.PvcQos.SendingFlowspec;

    // Initialize Qos Settings with default values : UBR without specified Qos

   	pRecvQos->TokenRate           = QOS_NOT_SPECIFIED;
	pRecvQos->TokenBucketSize     = QOS_NOT_SPECIFIED;
	pRecvQos->PeakBandwidth       = QOS_NOT_SPECIFIED;
	pRecvQos->Latency             = QOS_NOT_SPECIFIED;
	pRecvQos->DelayVariation      = QOS_NOT_SPECIFIED;
	pRecvQos->ServiceType         = SERVICETYPE_BESTEFFORT;
	pRecvQos->MaxSduSize          = QOS_NOT_SPECIFIED;
	pRecvQos->MinimumPolicedSize  = QOS_NOT_SPECIFIED;

    pSendQos->TokenRate           = QOS_NOT_SPECIFIED;
	pSendQos->TokenBucketSize     = QOS_NOT_SPECIFIED;
	pSendQos->PeakBandwidth       = QOS_NOT_SPECIFIED;
	pSendQos->Latency             = QOS_NOT_SPECIFIED;
	pSendQos->DelayVariation      = QOS_NOT_SPECIFIED;
	pSendQos->ServiceType         = SERVICETYPE_BESTEFFORT;
    pSendQos->MaxSduSize          = PacketSize;
	pSendQos->MinimumPolicedSize  = QOS_NOT_SPECIFIED;

	
    if ( QosType == QOS_UBR && SetQos)
        pSendQos->PeakBandwidth = pQosSettings->Ubr.PeakCellRate*48;

    else if (QosType == QOS_CBR)
    {
        pSendQos->ServiceType     = SERVICETYPE_GUARANTEED;
        pRecvQos->ServiceType     = SERVICETYPE_GUARANTEED;
        if (SetQos)
            pSendQos->PeakBandwidth = pQosSettings->Cbr.PeakCellRate*48;

    }
    else if (QosType == QOS_VBR)
    {
        pSendQos->ServiceType     = SERVICETYPE_CONTROLLEDLOAD;
        pRecvQos->ServiceType     = SERVICETYPE_CONTROLLEDLOAD;
        if (SetQos)
        {
            pSendQos->PeakBandwidth   = pQosSettings->Vbr.PeakCellRate*48;
            pSendQos->TokenRate       = pQosSettings->Vbr.SubstainableCellRate*48;
            pSendQos->TokenBucketSize = pQosSettings->Vbr.MaxBurstSize*48;
        }
    }


	PvcParameters.PvcConnectionId.VPI          = Vpi;
	PvcParameters.PvcConnectionId.VCI          = Vci;
	PvcParameters.PvcConnectionId.DeviceNumber = DeviceNumber;

	PvcParameters.PvcQos.ProviderSpecific.buf = NULL;
	PvcParameters.PvcQos.ProviderSpecific.len = 0;

	if (WSAIoctl(Socket, SIO_ASSOCIATE_PVC, (LPVOID) &PvcParameters,
				sizeof(ATM_PVC_PARAMS), NULL, 0, &OutBytes, NULL, NULL)
				== SOCKET_ERROR)
	{

		THROW_MESSAGE("Error: WSAIoctl(SIO_ASSOCIATE_PVC..). Error Code:%d", WSAGetLastError());
		closesocket(Socket);
	    WSACleanup();
		Socket = SOCKET_ERROR;
	}

    ConnectionInfo.Vpi =Vpi;
    ConnectionInfo.Vci =Vci;
    ConnectionInfo.DeviceNumber= DeviceNumber;
	ConnectionInfo.AALLayer = AalType;
	ConnectionInfo.Gfc      = Gfc;
	ConnectionInfo.Pt       = Pt;
	ConnectionInfo.Clp      = Clp;
   
    ConnectionInfo.PacketSize = PacketSize;	
	return (int) Socket;
}


// Starts the Thread that waits for the incomming connection
static BOOL StartWaitingCall(void)
{
    ConnectionInfo.WaitCallCloseEvent = WSACreateEvent();
   
    if (ConnectionInfo.WaitCallCloseEvent == NULL)
    {
   		THROW_MESSAGE("Error: WSACreateEvent. ErrorCode:%d", GetLastError());
        return FALSE;
    }
 
    
  	if ((ConnectionInfo.WaitCallThreadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WaitCallThread, 0, 0, &ConnectionInfo.WaitCallThreadID)) == 0)
    {
   
   		THROW_MESSAGE("Error: CreateThread(WaitCallThread ..). ErrorCode:%d", GetLastError());
        return FALSE;
    }

    return TRUE;

 
 

}

// Stops the Waiting Call Thread
static void StopWaitCall()
{
    if (ConnectionInfo.WaitCallThreadHandle && ConnectionInfo.WaitCallCloseEvent)
    {
        SetEvent(ConnectionInfo.WaitCallCloseEvent);
        
         // Wait 5 seconds for it to exit.  Shouldn't happen.
        if (WaitForSingleObject(ConnectionInfo.WaitCallThreadHandle, 5000) == WAIT_TIMEOUT)
        {
            TerminateThread(ConnectionInfo.WaitCallThreadHandle, 0);
        }
        

    }
    if (ConnectionInfo.WaitCallCloseEvent)
    {
        CloseHandle(ConnectionInfo.WaitCallCloseEvent);
        ConnectionInfo.WaitCallCloseEvent = 0;
    }

}

// Starts the  Receive Thread
static BOOL StartReceiveMode(void)
{
    
	int i;
    ConnectionInfo.RecvMode   = TRUE;    

    
    ConnectionInfo.RecvCloseEvent = WSACreateEvent();
    if (ConnectionInfo.RecvCloseEvent == NULL)
    {
   		THROW_MESSAGE("Error: WSACreateEvent. ErrorCode:%d", GetLastError());
        return FALSE;
    }
 

   
	for (i = 0; i <MAX_RECV_PACKETS; i++)
	{
		ConnectionInfo.RecvOverlapped[i].hEvent=WSACreateEvent();
		if (ConnectionInfo.RecvOverlapped[i].hEvent == NULL)
		{   
   		    THROW_MESSAGE("Error: WSACreateEvent. ErrorCode:%d", GetLastError());
            return FALSE;
		}

		if ((ConnectionInfo.RecvPacket[i].buf = (char *)malloc(ConnectionInfo.PacketSize) ) == NULL)
		{
   			THROW_MESSAGE("Error: malloc", 0);
			return FALSE;
		}

	}



  	if ((ConnectionInfo.RecvThreadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)RecvThread, 0, 0, &ConnectionInfo.RecvThreadID)) == 0)
    {
   
   		THROW_MESSAGE("Error: CreateThread(RecvThread ..). ErrorCode:%d", GetLastError());
        return FALSE;
    }
//	SetThreadPriority( ConnectionInfo.RecvThreadHandle, THREAD_PRIORITY_TIME_CRITICAL );
    return TRUE;


}

// Stops the Receive Thread
static void StopReceiveMode()
{
	int i;
    if (ConnectionInfo.RecvThreadHandle && ConnectionInfo.RecvCloseEvent)
    {
        SetEvent(ConnectionInfo.RecvCloseEvent);
        
         // Wait 5 seconds for it to exit.  Shouldn't happen.
        if (WaitForSingleObject(ConnectionInfo.RecvThreadHandle, 5000) == WAIT_TIMEOUT)
        {
            TerminateThread(ConnectionInfo.RecvThreadHandle, 0);
        }

    }


	for (i=0; i <MAX_RECV_PACKETS; i++)
	{

		if (ConnectionInfo.RecvOverlapped[i].hEvent)
		{
			CloseHandle(ConnectionInfo.RecvOverlapped[i].hEvent);
			ConnectionInfo.RecvOverlapped[i].hEvent = 0;
		}

		if (ConnectionInfo.RecvPacket[i].buf)
		{
            free(ConnectionInfo.RecvPacket[i].buf);    
            ConnectionInfo.RecvPacket[i].buf = NULL;
		}
	}

   
    if (ConnectionInfo.RecvCloseEvent)
    {
        CloseHandle(ConnectionInfo.RecvCloseEvent);
        ConnectionInfo.RecvCloseEvent = 0;
    }

}

// Starts the Transmit thread
static BOOL StartXmitMode(void)
{
    
    
    ConnectionInfo.XmitMode   = TRUE;    
 
    if ((ConnectionInfo.XmitPacket[0].buf = (char *)malloc(ConnectionInfo.PacketSize) ) == NULL)
    {
	    THROW_MESSAGE("Error: malloc", 0);
        return FALSE;
    }
    ConnectionInfo.XmitPacket[0].len = ConnectionInfo.PacketSize;   

    ConnectionInfo.XmitCloseEvent = WSACreateEvent();
    if (ConnectionInfo.XmitCloseEvent == NULL)
    {
        
   		THROW_MESSAGE("Error: WSACreateEvent. ErrorCode:%d", GetLastError());
        return FALSE;
    }
 
    
    ConnectionInfo.XmitOverlapped[0].hEvent=WSACreateEvent();
    if (ConnectionInfo.XmitOverlapped[0].hEvent == NULL)
    {   
   		THROW_MESSAGE("Error: WSACreateEvent. ErrorCode:%d", GetLastError());
        return FALSE;
    }


  	if ((ConnectionInfo.XmitThreadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)XmitThread, 0, 0, &ConnectionInfo.XmitThreadID)) == 0)
    {
   		THROW_MESSAGE("Error: CreateThread(XmitThread ..). ErrorCode:%d", GetLastError());
        return FALSE;
    }

    return TRUE;

}

// Stops the Transmit Thread
static void StopXmitMode()
{
    if (ConnectionInfo.XmitThreadHandle && ConnectionInfo.XmitCloseEvent)
    {
        SetEvent(ConnectionInfo.XmitCloseEvent);
        
         // Wait 5 seconds for it to exit.  Shouldn't happen.
        if (WaitForSingleObject(ConnectionInfo.XmitThreadHandle, 5000) == WAIT_TIMEOUT)
        {
            TerminateThread(ConnectionInfo.XmitThreadHandle, 0);
        }

    }

    if (ConnectionInfo.XmitOverlapped[0].hEvent)
    {
        CloseHandle(ConnectionInfo.XmitOverlapped[0].hEvent);
        ConnectionInfo.XmitOverlapped[0].hEvent = 0;
    }

    if (ConnectionInfo.XmitPacket[0].buf)
    {
            free(ConnectionInfo.XmitPacket[0].buf);    
            ConnectionInfo.XmitPacket[0].buf = NULL;
    }

      
    if (ConnectionInfo.XmitCloseEvent)
    {
        CloseHandle(ConnectionInfo.XmitCloseEvent);
        ConnectionInfo.XmitCloseEvent = 0;
    }
 
}


// The Thread that Waits for Incoming Call
static DWORD WaitCallThread(DWORD param)
{

    SOCKET NewSocket;
	struct sockaddr_atm SocketAddress;
	int    AddressLength;
    HANDLE HandleList[2];
    int ReturnCode;
    ATM_CONNECTION_ID ConnID;

    HandleList[0]= ConnectionInfo.hAccept;
    HandleList[1]= ConnectionInfo.WaitCallCloseEvent;
        
    ReturnCode=WaitForMultipleObjects(2,HandleList,FALSE,INFINITE);     // check state of ConnectionInfo.hAccept
    
    if (ReturnCode == WAIT_OBJECT_0 + 1 )
    {
    // User Stops WaitCAllThread
        if (ConnectionInfo.DialogHandle)
		    PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
 		    return SOCKET_ERROR;
    }

    else if (ReturnCode  != WAIT_OBJECT_0 ) 
    {
    // Unexpected end
        THROW_MESSAGE("Error: WaitForMultipleObjects. Error Code:%d", GetLastError());

        if (ConnectionInfo.DialogHandle)
		    PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
 		    return SOCKET_ERROR;


    }
    // Received Connect Event 
    CloseHandle(ConnectionInfo.WaitCallCloseEvent);
    ConnectionInfo.WaitCallCloseEvent = 0;

    memset((void *)&SocketAddress, 0, sizeof(struct sockaddr_atm));
	SocketAddress.satm_family = AF_ATM;
	SocketAddress.satm_number.AddressType = ATM_NSAP;
	SocketAddress.satm_number.NumofDigits = 20;
	SocketAddress.satm_blli.Layer2Protocol = SAP_FIELD_ANY; 
	SocketAddress.satm_blli.Layer3Protocol = SAP_FIELD_ABSENT;
	SocketAddress.satm_bhli.HighLayerInfoType = SAP_FIELD_ABSENT;
    
    AddressLength=sizeof(struct sockaddr_atm);
    if ((NewSocket = WSAAccept((SOCKET)ConnectionInfo.Socket, (struct sockaddr FAR *)&SocketAddress,&AddressLength,NULL,0)) 
		          == SOCKET_ERROR) 
    {
        THROW_MESSAGE("Error: accept. Error Code:%d", WSAGetLastError());
        if (ConnectionInfo.DialogHandle)
		    PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
 		    return SOCKET_ERROR;
    }
 
    memcpy(ConnectionInfo.RemoteATMAddress,SocketAddress.satm_number.Addr,20);
    
    ConnID.DeviceNumber = ConnectionInfo.DeviceNumber;
    if (GetATMConnectionId(NewSocket,&ConnID))
    {
        ConnectionInfo.Vpi=ConnID.VPI;
        ConnectionInfo.Vci=ConnID.VCI;

    }
    
    if (ConnectionInfo.DialogHandle)
        PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_CONNECTED, (LPARAM)NULL);

    ConnectionInfo.Socket=(HANDLE)NewSocket;

    if (ConnectionInfo.RecvMode)
    {
        if (StartReceiveMode() == FALSE)
        {
            SockStopATMConnection();
            return STATE_NOT_CONNECTED;

        }


    }
    if (ConnectionInfo.XmitMode)
    {
        if (StartXmitMode() == FALSE)
        {
            SockStopATMConnection();
            return STATE_NOT_CONNECTED;
        }
    }
    
    return SUCCESS;
}


// The Receive Thread
static DWORD RecvThread(DWORD param)
{
	DWORD  bytesRecv, RecvFlags;
	PPACKET_HEADER pPacketHeader;
    int i,iError;
	PAAL0_CELL pCell;
	DWORD dwEvent;

	int CurrentIndex=0;
	

	struct timeval timeselect={1,0};

    ConnectionInfo.RecvSeq =0;
	iError = 0;	

	for (i=1; i < MAX_RECV_PACKETS; i++)
	{
		RecvFlags = 0 ;
		ConnectionInfo.RecvPacket[i].len = ConnectionInfo.PacketSize;
		if ( WSARecv((SOCKET)ConnectionInfo.Socket, &ConnectionInfo.RecvPacket[i],1, NULL, &RecvFlags,
				                &ConnectionInfo.RecvOverlapped[i], NULL) != 0)
		{
			iError =WSAGetLastError();
			if (iError != WSA_IO_PENDING)
				break;

			ConnectionInfo.hRecvEventList[i]=ConnectionInfo.RecvOverlapped[i].hEvent;
		}
	}
    ConnectionInfo.hRecvEventList[MAX_RECV_PACKETS]=ConnectionInfo.RecvCloseEvent;

	if (iError != 0 && iError != WSA_IO_PENDING)
	{
		THROW_MESSAGE("Error: WSARecv. ErrorCode:%d", iError);
		if (ConnectionInfo.DialogHandle)
			PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
	}
	else
	{
		while(TRUE)
		{
			dwEvent = WSAWaitForMultipleEvents(1 + MAX_RECV_PACKETS - CurrentIndex,
				                         &ConnectionInfo.hRecvEventList[CurrentIndex] , FALSE, WSA_INFINITE,FALSE);


			if (dwEvent <  (WAIT_OBJECT_0 + MAX_RECV_PACKETS - CurrentIndex))
			{ 
				CurrentIndex +=dwEvent;
//				WSAResetEvent(ConnectionInfo.RecvOverlapped[CurrentIndex].hEvent);
				while (WSAGetOverlappedResult((SOCKET)ConnectionInfo.Socket,&ConnectionInfo.RecvOverlapped[CurrentIndex], &bytesRecv,FALSE,&RecvFlags) == TRUE)
				{
					if (ConnectionInfo.AALLayer != LAYER_AAL0)
						pPacketHeader = (PPACKET_HEADER) ConnectionInfo.RecvPacket[CurrentIndex].buf;
					else
					{
						pCell = (PAAL0_CELL)ConnectionInfo.RecvPacket[CurrentIndex].buf;
						ConnectionInfo.RecvCellHeader = ntohl(pCell->CellHeader);
						pPacketHeader = (PPACKET_HEADER)pCell->Payload;
					}
						
					if (pPacketHeader->seq != ConnectionInfo.RecvSeq)
					{
						if (ConnectionInfo.RecvSeq)
							ConnectionInfo.pktDrop += (pPacketHeader->seq - ConnectionInfo.RecvSeq);
					}
					ConnectionInfo.RecvSeq = pPacketHeader->seq + 1;
					ConnectionInfo.pktRecv ++;
					ConnectionInfo.bytesRecv += bytesRecv;


					if (WSARecv((SOCKET)ConnectionInfo.Socket, &ConnectionInfo.RecvPacket[CurrentIndex],1, NULL, &RecvFlags,
								&ConnectionInfo.RecvOverlapped[CurrentIndex], NULL) != 0)
					{
						iError =WSAGetLastError();
						if (iError != WSA_IO_PENDING)
							break;
					}
					CurrentIndex = (CurrentIndex +1) %MAX_RECV_PACKETS;

				}
				if ( ((iError = WSAGetLastError()) !=WSA_IO_INCOMPLETE) && (iError != WSA_IO_PENDING))
				{
					THROW_MESSAGE("Error: WSAGetOverlappedResult. ErrorCode:%d", WSAGetLastError());
					if (ConnectionInfo.DialogHandle)
       					PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
					break;
				}
	        
			}
			else if (dwEvent ==  WAIT_OBJECT_0 + MAX_RECV_PACKETS - CurrentIndex )
			{
				// close event
				break;
			}

			else
			{
				THROW_MESSAGE("Error: WSAWaitForMultipleEvents. ErrorCode:%d", WSAGetLastError());
				if (ConnectionInfo.DialogHandle)
       				PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
				break;
	        
			}
		}
	}
	ConnectionInfo.RecvThreadHandle = NULL;
    return SUCCESS;
}

// The Transmit Thread
static DWORD XmitThread(DWORD param)
{
	DWORD  bytesXmit, XmitFlags;
	PPACKET_HEADER pPacketHeader;
	PAAL0_CELL pCell;
    int iError;
	DWORD dwEvent;

	struct timeval timeselect={1,0};

    ConnectionInfo.hXmitEventList[0]=ConnectionInfo.XmitCloseEvent;
    ConnectionInfo.hXmitEventList[1]=ConnectionInfo.XmitOverlapped[0].hEvent;
	if (ConnectionInfo.AALLayer != LAYER_AAL0)
	{
		pPacketHeader = (PPACKET_HEADER) ConnectionInfo.XmitPacket[0].buf;
		pPacketHeader->seq = 0;
		pPacketHeader->len = ConnectionInfo.XmitPacket[0].len;
	
	}
	else

	{
		pCell = (PAAL0_CELL)ConnectionInfo.XmitPacket[0].buf;

		pCell->CellHeader =     (((ConnectionInfo.Gfc << 28) & 0xF0000000) |  
			                    ((ConnectionInfo.Vpi << 20) & 0x0FF00000) | 
						        ((ConnectionInfo.Vci << 4 ) & 0x000FFFF0) | 
						        ((ConnectionInfo.Pt  << 1)  & 0x0000000E) | 
						        (ConnectionInfo.Clp         & 0x00000001));
		pCell->CellHeader = htonl(pCell->CellHeader);

//$$$$$$$$$$ for OAM test puposes $$$$$$$$$$$$$
//		pCell->Payload[0]=0x81;
//	    pCell->Payload[1]=0x5;
//		memset(&pCell->Payload[2], 0,46);
//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

		pPacketHeader = (PPACKET_HEADER)pCell->Payload;
		pPacketHeader->len = 52;
		pPacketHeader->seq =0;
	}

    while(TRUE)
	{	
		
		XmitFlags = ZERO_FLAGS;
        bytesXmit=ConnectionInfo.XmitPacket[0].len;

	    if (WSASend((SOCKET)ConnectionInfo.Socket, &ConnectionInfo.XmitPacket[0],1, &bytesXmit, XmitFlags,
                    &ConnectionInfo.XmitOverlapped[0], NULL) != 0)
		{

            iError = WSAGetLastError();
            if (iError != WSA_IO_PENDING)
            {
                THROW_MESSAGE("Error: WSASend. ErrorCode:%d", iError);
                if (ConnectionInfo.DialogHandle)
  			            PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
                break;
            }
            else
            {
                
         		dwEvent = WSAWaitForMultipleEvents(2,ConnectionInfo.hXmitEventList , FALSE, WSA_INFINITE,FALSE);
        
		        if (dwEvent ==  WAIT_OBJECT_0)
		        {
		            // close event 
			        break;
		        }

		        else if (dwEvent <=  (WAIT_OBJECT_0 + 2 ))
		        { 
                    if (WSAGetOverlappedResult((SOCKET)ConnectionInfo.Socket,&ConnectionInfo.XmitOverlapped[0],
                        &bytesXmit,TRUE,&XmitFlags) != TRUE)
                    {

                        THROW_MESSAGE("Error: WSAGetOverlappedResult. ErrorCode:%d", WSAGetLastError());
                        if (ConnectionInfo.DialogHandle)
       			            PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
                        break;
                    }
                
                }
                else
                {
                    THROW_MESSAGE("Error: WSAWaitForMultipleEvents. ErrorCode:%d", WSAGetLastError());
                    if (ConnectionInfo.DialogHandle)
  			            PostMessage(ConnectionInfo.DialogHandle,PWM_INF_MESSAGE, STATE_NOT_CONNECTED, (LPARAM)NULL);
                    break;
                }


            }
		}

			
		pPacketHeader->seq ++;			
		ConnectionInfo.pktXmit = pPacketHeader->seq;
		ConnectionInfo.bytesXmit += bytesXmit;


   		dwEvent = WSAWaitForMultipleEvents(1,ConnectionInfo.hXmitEventList , FALSE, 0,FALSE);
        
        if (dwEvent ==  WAIT_OBJECT_0)
        {
		            // close event
	        break;
        }
		if (bStepMode)
			if (AfxMessageBox(" One packet send . Press YES to resend or NO to abort", MB_YESNO |MB_USERICON ) == IDNO)
				break;


 
	}
	
    ConnectionInfo.XmitThreadHandle = NULL;
	return SUCCESS;
}



// Retreives the ATM Address when the adapter is connected to a switch
BOOL SockGetLocalATMAddress(int DeviceNumber,BYTE *pLocalATMAddress)
{

	SOCKET sock;
	ATM_ADDRESS addr;
	DWORD bytes;
    
    if (!InitWS2())
        return FALSE;
    	
	if ((sock = WSASocket(AF_ATM, SOCK_RAW, ATMPROTO_AAL5, NULL, 0, 0))	== INVALID_SOCKET) 
    {
		THROW_MESSAGE("Error: WSASocket. Error Code:%d", WSAGetLastError());
		return FALSE;
    }

	if (WSAIoctl(sock, SIO_GET_ATM_ADDRESS, (LPVOID) &DeviceNumber, sizeof(DWORD), (LPVOID) &addr,
		         sizeof(ATM_ADDRESS), &bytes, NULL, NULL) == SOCKET_ERROR)
	{
		THROW_MESSAGE("Error: WSAIoctl(SIO_GET_ATM_ADDRESS..). Error Code:%d", WSAGetLastError());
        return FALSE;
	}

	closesocket(sock);
	WSACleanup();

    memcpy(pLocalATMAddress,addr.Addr,20);

    return TRUE;
 
}

// 


// Retreives the Local ATM Device Count 
BOOL SockGetATMDevices( PATM_ADAPTER pAdapters, DWORD  *pAdapterCount)
{
	SOCKET sock;
	DWORD bytes;
	DWORD AdapterCount, DeviceCount;
   
	AdapterCount = *pAdapterCount;
	FindNetworkATMAdapters( pAdapters, &AdapterCount);
	if (!AdapterCount)
		return FALSE;

	if (!InitWS2())
	{
		return FALSE;
	}
	if ((sock = WSASocket(AF_ATM, SOCK_RAW, ATMPROTO_AAL5, NULL, 0, 0))	== INVALID_SOCKET) 
	{
		THROW_MESSAGE("Error: WSASocket. Error Code:%d", WSAGetLastError());
		return FALSE;
	}

	/* Check available ATM devices */
	if(WSAIoctl(sock,
            SIO_GET_NUMBER_OF_ATM_DEVICES,
            NULL,
            0,
            &DeviceCount,
            sizeof(DWORD),
            &bytes,
            NULL,
            NULL) != 0)
	{

		THROW_MESSAGE("Error: WSAIoctl(SIO_GET_NUMBER_OF_ATM_DEVICES..). Error Code:%d", WSAGetLastError());
		return FALSE;

	}

	closesocket(sock);

	WSACleanup();
	*pAdapterCount = min( AdapterCount, DeviceCount); 
	return TRUE;

}



// Starts the ATM Connection according to the pSettings Contents
int SockStartATMConnection(PATM_SETTINGS pSettings,HWND DialogHandle)
{
    
    memset(&ConnectionInfo,0,sizeof(CONNECTION_INFO));

    ConnectionInfo.DialogHandle =DialogHandle;

    if (pSettings->VcType == CONNECTION_PVC)
    {

                
        if ((ConnectionInfo.Socket=(HANDLE)CreatePVC(pSettings->DeviceNumber,
                                         pSettings->VcSettings.Pvc.Vpi,
                                         pSettings->VcSettings.Pvc.Vci,
                                         pSettings->AalType,
										 pSettings->VcSettings.Pvc.Gfc,
										 pSettings->VcSettings.Pvc.Pt,
										 pSettings->VcSettings.Pvc.Clp,
                                         pSettings->QualityOfServices,
                                         pSettings->SetQosSettings,
                                         &pSettings->QosSettings, 
                                         pSettings->PacketSize)) == (HANDLE)SOCKET_ERROR)
        {
            return STATE_NOT_CONNECTED;
            
        }
        ConnectionInfo.ConnectionState = STATE_CONNECTED;

    }
    else // Incomming SVC
    {
        if ((ConnectionInfo.Socket=(HANDLE)CreateSVC(pSettings->DeviceNumber,
                                         pSettings->VcType,
                                         &pSettings->VcSettings.Svc,
                                         pSettings->QualityOfServices,
                                         pSettings->SetQosSettings,
                                         &pSettings->QosSettings,
                                         pSettings->PacketSize )) == (HANDLE)SOCKET_ERROR)
        {
            return STATE_NOT_CONNECTED;
            
        }
        // When it's an incomming Connection, we have an intermediate state in which a thread waits for the caller
        if (pSettings->VcType == CONNECTION_SVC_IN)
            ConnectionInfo.ConnectionState =STATE_CONNECTING;
        else
            ConnectionInfo.ConnectionState = STATE_CONNECTED;

        
    }

    ConnectionInfo.VcType = pSettings->VcType;
    
    if (ConnectionInfo.ConnectionState == STATE_CONNECTING)
    {
        ConnectionInfo.RecvMode   = pSettings->ReceiveFlow;    
        ConnectionInfo.XmitMode   = pSettings->XmitFlow;    
        // Starts the waiting thread
        StartWaitingCall();

    }
    else
    {

        if (pSettings->ReceiveFlow)
        {
            if (StartReceiveMode() == FALSE)
            {
                SockStopATMConnection();
                return STATE_NOT_CONNECTED;

            }


        }
        if (pSettings->XmitFlow)
        {
            if (StartXmitMode() == FALSE)
            {
                SockStopATMConnection();
                return STATE_NOT_CONNECTED;
            }
        }
    }
    return ConnectionInfo.ConnectionState;
 }

// Stops the ATM Connection
void SockStopATMConnection()
{

    if (ConnectionInfo.ConnectionState == STATE_CONNECTING)
        StopWaitCall();

    if(ConnectionInfo.XmitMode)
        StopXmitMode();

    if (ConnectionInfo.RecvMode)
        StopReceiveMode();
            
    if (ConnectionInfo.Socket)
    {
        closesocket((SOCKET)ConnectionInfo.Socket);
      	WSACleanup();

    }

    memset(&ConnectionInfo,0,sizeof(CONNECTION_INFO));


}